/*
Copyright (c) 2001 Bob Trower, Trantor Standard Systems Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// Modified from original code (http://base64.sourceforge.net/) for in-memory encoding and decoding.
// The decoding code has been moved into LUA_FUNCTION(b64_decode) in gm_zlib_b64.cpp
#include "b64.h"


// Translation Table as described in RFC1113
static const char cb64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*
** encodeblock
** encode 3 8-bit binary bytes as 4 '6-bit' characters
*/
void encodeblock(unsigned char in[3], unsigned char* out, int len)
{
	out[0] = cb64[in[0] >> 2];
	out[1] = cb64[((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4)];
	out[2] = (unsigned char)(len > 1 ? cb64[((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6)] : '=');
	out[3] = (unsigned char)(len > 2 ? cb64[in[2] & 0x3f] : '=');
}


/*
** encode
** base64 encode a buffer.
** 'in' parameter allows encoding of multiple buffers, pass the 'in' parameter
** of the previous call to the next call.
** To end the encoding, pass the 'in' parameter to the 'encodeblock' function.
** Returns false on buffer overrun, true on completion.
*/
void encode(ENCODE_DATA* data, const unsigned char* src, size_t src_len)
{
	const unsigned char* srctmp = src;

	while (srctmp < (src + src_len)) {
		for (int i = data->len; i < 3; i++) {
			data->in[i] = *srctmp;
			srctmp++;
			if (srctmp <= (src + src_len)) data->len++; else data->in[i] = 0;
		}
		if (data->len == 3) {
			encodeblock(data->in, data->out, 3);
			data->out += 4;
			data->len = 0;
		} else return;
	}
}