/*
gm_zlib_b64.dll - Garry's Mod zlib compression and base64 encoding module
Copyright (C) 2008 Ieuan Henry

This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General
Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option)
any later version.

This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to
the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
Declarations
#define, #include, structure, globals, etc.
*/
#define META_COMPBUFFER "CompBuffer"	// Type name
#define TYPE_COMPBUFFER 2953			// Random type ID
#define ZLIB_CHECKSUM(_cmp_) (*((unsigned long*)(_cmp_->data + _cmp_->comp_size - 4)))

#include "common/GMLuaModule.h"
ILuaModuleManager* modulemanager = NULL;

#include "b64.h"			// base64 stuff
extern "C" {
#include "zlib/zutil.h"		// zlib header file (with internal functions)
}
#include <windows.h>		// Windows APIs

// The CompBuffer Lua type - contains compressed data and size, and uncompressed size.
struct CompBuffer
{
	unsigned short header_size;
	char* header;		// header data added using SetHeader, default = NULL, NULL terminated in files
	unsigned long uncomp_size;
	unsigned char* data;
	unsigned long comp_size;
};
HANDLE hHeap;

// Base64: Translation Table to decode.
static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";



// CreateFileGMod - CreateFile that takes GMod relative paths.
HANDLE CreateFileGMod(LPCSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile)
{
	char path[MAX_PATH];	
	strcpy(path, modulemanager->GetBaseFolder());
	strcat(path, "data\\");
	strcat(path, lpFileName);

	// 'path' now contains the full path to the file - open it.
	// CreateFile accepts forward slashes ('/'), so no need to replace them (if any).
	HANDLE f = CreateFile(path, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile);
	return f;
}
//


/*
zlib functions:
	zlib_compress = zlib.Compress
	zlib_decompress = zlib.Decompress, CompBuffer:Decompress
zlib is Copyright (c) 1995-2004 Jean-loup Gailly and Mark Adler. See zlib/README for details.
*/
// CompBuffer zlib.Compress(string src, [number complevel], [number strategy])
LUA_FUNCTION(zlib_compress)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, GLua::TYPE_STRING);

	const char* data = gLua->GetString(1);
	CompBuffer* cmp = (CompBuffer*)HeapAlloc(hHeap, 0, sizeof(CompBuffer));

	// Optional compression level parameter
	int complevel = 6, strategy = Z_DEFAULT_STRATEGY;
	if (gLua->GetType(2) == GLua::TYPE_NUMBER) complevel = gLua->GetInteger(2);
	if (gLua->GetType(3) == GLua::TYPE_NUMBER) strategy = gLua->GetInteger(3);

	cmp->uncomp_size = gLua->StringLength(1);
	// minimum destination size (http://www.zlib.net/manual.html#compress), +1 in case of rounding errors.
	cmp->comp_size = cmp->uncomp_size + (cmp->uncomp_size / 1000) + 13;
	cmp->data = (unsigned char*)HeapAlloc(hHeap, 0, cmp->comp_size);

	// compress the data
	// --- copied from 'compress2' function and edited to fit here ---
	z_stream stream = {
		// next_in, avail_in, total_in
		(Bytef*)data, cmp->uncomp_size, 0,
		// next_out, avail_out, total_out
		cmp->data, cmp->comp_size, 0,
		// msg, internal_state
		NULL, NULL,
		// zalloc, zfree, opaque
		(alloc_func)0, (free_func)0, hHeap,
		// data_type (will always be Z_TEXT for this module), adler, reserved
		Z_TEXT, 0, 0
	};

	// @!:: changed to deflateInit2 to use strategy parameter
	int err = deflateInit2(&stream, complevel, Z_DEFLATED, 15, 9, strategy);
	if (err != Z_OK) goto zlib_err;

	err = deflate(&stream, Z_FINISH);
	if (err != Z_STREAM_END) {
		deflateEnd(&stream);
		err = (err == Z_OK ? Z_BUF_ERROR : err);
		goto zlib_err;
	}
	cmp->comp_size = stream.total_out;

	err = deflateEnd(&stream);
	if (err != Z_OK) goto zlib_err;
	// -------------------------------------------------------------

	// Free unused memory (leave 4 bytes for uncompressed size)
	cmp->data = (unsigned char*)HeapReAlloc(hHeap, NULL, cmp->data, cmp->comp_size);
	// Zero header fields
	cmp->header = NULL;
	cmp->header_size = 0;

	// Set metatable and return
	ILuaObject* CompBufferMT = gLua->GetMetaTable(META_COMPBUFFER, TYPE_COMPBUFFER);
	gLua->PushUserData(CompBufferMT, (void*)cmp);		// Return the CompBuffer
	CompBufferMT->UnReference();
	return 1;
zlib_err:
	HeapFree(hHeap, 0, cmp->data);
	HeapFree(hHeap, 0, cmp);
	gLua->Error(zError(err));
	return 0;
}
// string zlib.Decompress(CompBuffer cmp)
LUA_FUNCTION(zlib_decompress)
{
	ILuaInterface* gLua = Lua();

	gLua->CheckType(1, TYPE_COMPBUFFER);
	CompBuffer* cmp = (CompBuffer*)gLua->GetUserData(1);
	char* dest = (char*)HeapAlloc(hHeap, NULL, cmp->uncomp_size + 1);		// +1 for null termination
	
	// --- copied from uncompr.c, edited to fit in here ---
	z_stream stream = {
		// next_in, avail_in, total_in
		cmp->data, cmp->comp_size, 0,
		// next_out, avail_out, total_out
		(Bytef*)dest, cmp->uncomp_size, 0,
		// msg, internal_state
		NULL, NULL,
		// zalloc, zfree, opaque
		(alloc_func)0, (free_func)0, hHeap,
		// data_type (will always be Z_TEXT for this module), adler, reserved
		Z_TEXT, 0, 0
	};

	int err = inflateInit2(&stream, 15);
	if (err != Z_OK) return err;

	err = inflate(&stream, Z_FINISH);
	if (err != Z_STREAM_END) {
		inflateEnd(&stream);
		err = ((err == Z_NEED_DICT) || ((err == Z_BUF_ERROR) && (stream.avail_in == 0))) ? Z_DATA_ERROR : err;
		goto zlib_err;
	}
	if (stream.total_out != cmp->uncomp_size) {
		inflateEnd(&stream);
		err = Z_BUF_ERROR;
		goto zlib_err;
	}

	err = inflateEnd(&stream);
	if (err != Z_OK) goto zlib_err;
	// ----------------------------------------------------

	dest[cmp->uncomp_size] = 0;	// null-terminate output
	gLua->Push(dest);
	// GMod makes a copy of pushed strings, so the memory allocated is no longer needed - free it.
	HeapFree(hHeap, NULL, dest);
	return 1;
zlib_err:
	HeapFree(hHeap, 0, dest);
	gLua->Error(zError(err));
	return 0;
}
//


/*
CompBuffer functions:
	compbuffer_gc = CompBuffer:__gc
	compbuffer_len = CompBuffer:__len, CompBuffer:CompressedSize
	compbuffer_eq = CompBuffer:__eq
	compbuffer_uncomp_size = CompBuffer:UncompressedSize
	compbuffer_write = CompBuffer:Write
-- New for version 1.1 --
	compbuffer_setheader = CompBuffer:SetHeader
	compbuffer_getheader = CompBuffer:GetHeader
-------------------------
	compbuffer_read = zlib.Read
*/
// void CompBuffer:__gc()
LUA_FUNCTION(compbuffer_gc)
{
	if (hHeap == NULL) return 0;		// Fix for disconnect crash bug
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);

	CompBuffer* cmp = (CompBuffer*)gLua->GetUserData(1);
	HeapFree(hHeap, 0, cmp->data);
	if (cmp->header) HeapFree(hHeap, 0, cmp->header);
	HeapFree(hHeap, 0, cmp);
	return 0;
}
// number CompBuffer:__len(); number CompBuffer
LUA_FUNCTION(compbuffer_len)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	gLua->Push((float)(((CompBuffer*)gLua->GetUserData(1))->comp_size));
	return 1;
}
// bool CompBuffer:__eq(CompBuffer cmp2)
LUA_FUNCTION(compbuffer_eq)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	gLua->CheckType(2, TYPE_COMPBUFFER);

	CompBuffer* cmp1 = (CompBuffer*)gLua->GetUserData(1);
	CompBuffer* cmp2 = (CompBuffer*)gLua->GetUserData(2);
	
	// check sizes and checksums
	if ((cmp1->comp_size != cmp2->comp_size) || (cmp1->uncomp_size != cmp2->uncomp_size) || (ZLIB_CHECKSUM(cmp1) != ZLIB_CHECKSUM(cmp2)) || (cmp1->header_size != cmp2->header_size)) goto returnfalse;

	// check headers
	if (((cmp1->header == 0) && (cmp2->header == 0)) || ((cmp1->header != 0) && (cmp2->header != 0))) {
		for (unsigned long i = 0; i < cmp1->header_size; i++) {
			if (cmp1->header[i] != cmp2->header[i]) goto returnfalse;
		}
	} else goto returnfalse;

	// check contents
	for (unsigned long i = 0; i < cmp1->comp_size; i++) {
		if (cmp1->data[i] != cmp2->data[i]) goto returnfalse;
	}

	// passed checks - return true
	gLua->Push(true);  return 1;
returnfalse:
	// failed a check - return false
	gLua->Push(false); return 1;
}
// number CompBuffer:UncompressedSize()
LUA_FUNCTION(compbuffer_uncomp_size)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	gLua->Push((float)(((CompBuffer*)gLua->GetUserData(1))->uncomp_size));
	return 1;
}
//
//-- new for version 1.1 --
// void CompBuffer:SetHeader(string header)
LUA_FUNCTION(compbuffer_setheader)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	CompBuffer* cmp = (CompBuffer*)gLua->GetUserData(1);

	if (gLua->GetType(2) == GLua::TYPE_NIL) goto removeheader;
	gLua->CheckType(2, GLua::TYPE_STRING);

	const char* header = gLua->GetString(2);
	unsigned long hs = gLua->StringLength(2);
	if (hs == 0) goto removeheader;
	if (hs > 65535) gLua->LuaError("The maximum size of a header is 65535 bytes.", 2);
	cmp->header_size = (unsigned short)hs;

	// If there is already a header, free it.
	if (cmp->header) HeapFree(hHeap, 0, cmp->header);
	cmp->header = (char*)HeapAlloc(hHeap, 0, cmp->header_size + 1);
	strcpy(cmp->header, header);
	return 0;

removeheader:
	cmp->header_size = 0;
	if (cmp->header) HeapFree(hHeap, 0, cmp->header);
	cmp->header = NULL;
	return 0;
}
// void CompBuffer:GetHeader()
LUA_FUNCTION(compbuffer_getheader)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	gLua->Push(((CompBuffer*)gLua->GetUserData(1))->header);
	return 1;
}
//--------------------------
//`
// void CompBuffer:Write(string path)
LUA_FUNCTION(compbuffer_write)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);
	gLua->CheckType(2, GLua::TYPE_STRING);

	const char* path = gLua->GetString(2);
	// Security checks: Find '..'s and check the extension is '.z'
	unsigned long i;
	for (i = 0; path[i] != 0; i++) {
		if ((path[i] == '.') && (path[i+1] == '.')) gLua->LuaError("'..' in path - cannot write outside data directory", 2);
	}
	if ((path[i-2] != '.') || (path[i-1] != 'z')) gLua->LuaError("The file must have a '.z' extension", 2);

	HANDLE f = CreateFileGMod(path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (f == NULL) {	// file I/O error
		char errmsg[128];
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), NULL, errmsg, 128, NULL);
		gLua->LuaError(errmsg, 2);
		return 0;
	}

	CompBuffer* cmp = (CompBuffer*)gLua->GetUserData(1);
	char c = (cmp->header ? 1 : 0);
	unsigned long __temp = 0;

	WriteFile(f, &c, 1, &__temp, NULL);
	// write header
	if (cmp->header) {
		WriteFile(f, &cmp->header_size, 2, &__temp, NULL);
		WriteFile(f, cmp->header, cmp->header_size, &__temp, NULL);
	}
	// uncompressed size
	WriteFile(f, &cmp->uncomp_size, 4, &__temp, NULL);
	// data
	WriteFile(f, cmp->data, cmp->comp_size, &__temp, NULL);

	// close file and return
	FlushFileBuffers(f);
	CloseHandle(f);
	return 0;
}
// CompBuffer zlib.Read(string path)
LUA_FUNCTION(compbuffer_read)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, GLua::TYPE_STRING);

	HANDLE f = CreateFileGMod(gLua->GetString(1), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (f == INVALID_HANDLE_VALUE) {	// file I/O error
		char errmsg[128];
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), NULL, errmsg, 128, NULL);
		gLua->LuaError(errmsg, 1);
		return 0;
	}
	
	CompBuffer* cmp = (CompBuffer*)HeapAlloc(hHeap, NULL, sizeof(CompBuffer));
	cmp->comp_size = GetFileSize(f, NULL) - 1;
	
	// has a header?
	unsigned long __temp = 0;
	char c;
	ReadFile(f, &c, 1, &__temp, NULL);
	if (c == 1) {
		// has a header - read the header's size
		ReadFile(f, &cmp->header_size, 2, &__temp, NULL);
		// allocate memory and read the header
		cmp->header = (char*)HeapAlloc(hHeap, 0, cmp->header_size + 1);
		ReadFile(f, cmp->header, cmp->header_size, &__temp, NULL);
		cmp->header[cmp->header_size] = 0;
		// take header size from compressed size
		cmp->comp_size -= (cmp->header_size + 2);
	} else if (c == 0) {
		// no header.
		cmp->header = NULL;
		cmp->header_size = 0;
	} else 
		gLua->Error("gm_zlib_b64: Unknown header byte value.");

	// read the uncompressed size
	ReadFile(f, &cmp->uncomp_size, 4, &__temp, NULL);
	cmp->comp_size -= 4;

	// read compressed data
	cmp->data = (unsigned char*)HeapAlloc(hHeap, 0, cmp->comp_size);
	ReadFile(f, cmp->data, cmp->comp_size, &__temp, NULL);

	// check this is the end of the file - if not something went wrong.
	ReadFile(f, &c, 1, &__temp, NULL);
	if (__temp == 1) gLua->Error("gm_zlib_b64: Expected end of file, got extra data (bug?).");

	// clean up, set metatable and return.
	CloseHandle(f);
	ILuaObject* CompBufferMT = gLua->GetMetaTable(META_COMPBUFFER, TYPE_COMPBUFFER);
	gLua->PushUserData(CompBufferMT, (void*)cmp);		// Return the CompBuffer
	CompBufferMT->UnReference();
	return 1;
}
//


/*
Base64 functions:
	b64_encode = b64.Encode
	b64_decode = b64.Decode
See b64.h for details.
*/
// string b64.Encode(CompBuffer src, [bool encodeheader])
LUA_FUNCTION(b64_encode)
{
	struct ENCODE_DATA encdata;
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, TYPE_COMPBUFFER);

	CompBuffer* cmp = (CompBuffer*)gLua->GetUserData(1);
	// +5 for uncompressed size and header byte, +1 for null-termination, +4 for padding.
	size_t destSize = (((cmp->comp_size + 5 - ((cmp->comp_size + 5) % 3) + 3) / 3) * 4) + 1 + 4;
	// +2 for header size, +4 for padding.
	if (cmp->header) destSize += (((cmp->header_size + 2 - ((cmp->header_size + 2) % 3) + 3) / 3) * 4) + 4;

	encdata.dest = (unsigned char*)HeapAlloc(hHeap, 0, destSize);
	encdata.dest_len = destSize;
	encdata.in[0] = (cmp->header ? 1 : 0);		// 'has header?' byte
	encdata.len = 1;
	encdata.out = encdata.dest;

	if (cmp->header) {
		encode(&encdata, (const unsigned char*)&cmp->header_size, 2);			// header size
		encode(&encdata, (const unsigned char*)cmp->header, cmp->header_size);	// header
	}
	encode(&encdata, (const unsigned char*)&cmp->uncomp_size, 4);				// uncompressed size
	encode(&encdata, cmp->data, cmp->comp_size);		// data

	// finish encoding
	if (encdata.len > 0) {
		encodeblock(encdata.in, encdata.out, encdata.len);
		encdata.out += 4;
	}
	// null-terminate
	*encdata.out = 0;

	// return it
	gLua->Push((char*)encdata.dest);
	HeapFree(hHeap, 0, encdata.dest);
	return 1;
}
// CompBuffer b64.Decode(string src)
LUA_FUNCTION(b64_decode)
{
	ILuaInterface* gLua = Lua();
	gLua->CheckType(1, GLua::TYPE_STRING);

	const unsigned char* src = (const unsigned char*)gLua->GetString(1);
	unsigned long srcSize = gLua->StringLength(1);
	unsigned long destSize = (srcSize / 4) * 3;
	// allocate temporary space
	unsigned char* dest = (unsigned char*)HeapAlloc(hHeap, 0, destSize);
	unsigned char* desttmp = dest;

	// ----------------------------------------------
	// decode
	unsigned char* out = dest;
	const unsigned char* srctmp = (const unsigned char*)src;
	unsigned char in[4], v;
	int i, len;

	while (srctmp < (src + srcSize)) {
		for (len = 0, i = 0; i < 4 && (srctmp < (src + srcSize)); i++) {
			v = 0;
			while ((srctmp < (src + srcSize)) && v == 0) {
				v = *srctmp; srctmp++;
				v = (unsigned char)((v < 43 || v > 122) ? 0 : cd64[v - 43]);
				if (v) v = (unsigned char) ((v == '$') ? 0 : v - 61);
			}
			if (srctmp <= (src + srcSize)) {
				if (v) { in[i] = (unsigned char)(v - 1); len++; } else in[i] = 0;
			} else in[i] = 0;
		}
		if (len) {
			*out = (unsigned char)(in[0] << 2 | in[1] >> 4);
			*(++out) = (unsigned char)(in[1] << 4 | in[2] >> 2);
			*(++out) = (unsigned char)(((in[2] << 6) & 0xc0) | in[3]);
			out++;
		}
	}
	// ----------------------------------------------

	// allocate CompBuffer
	CompBuffer* cmp = (CompBuffer*)HeapAlloc(hHeap, NULL, sizeof(CompBuffer));

	if (dest[0] == 1) {
		// has a header
		cmp->header_size = *((unsigned short*)++desttmp);
		cmp->header = (char*)HeapAlloc(hHeap, 0, cmp->header_size + 1);
		desttmp += 2;
		zmemcpy((unsigned char*)cmp->header, desttmp, cmp->header_size);
		cmp->header[cmp->header_size] = 0;
		desttmp += cmp->header_size;
	} else if (dest[0] == 0) {
		// no header
		cmp->header = NULL;
		cmp->header_size = 0;
		desttmp++;
	} else gLua->Error("gm_zlib_b64: Unknown header byte value.");

	// get uncompressed size and data
	cmp->uncomp_size = *((unsigned long*)desttmp);
	desttmp += 4;
	// The size of the output, -1 for each '=' that the input is padded with.
	cmp->comp_size = dest + destSize - desttmp - ((*(srctmp - 1) == '=') ? ((*(srctmp - 2) == '=') ? 2 : 1) : 0);
	cmp->data = (unsigned char*)HeapAlloc(hHeap, 0, cmp->comp_size);

	zmemcpy(cmp->data, desttmp, cmp->comp_size);
	HeapFree(hHeap, 0, dest);

	// Set metatable and return
	ILuaObject* CompBufferMT = gLua->GetMetaTable(META_COMPBUFFER, TYPE_COMPBUFFER);
	gLua->PushUserData(CompBufferMT, (void*)cmp);		// Return the CompBuffer
	CompBufferMT->UnReference();
	return 1;
}
//


// Functions from GMOD_MODULE macro [done myself to save 2 function calls :)]
extern "C" int __declspec(dllexport) gmod_open(ILuaInterface* i)
{
	modulemanager = i->GetModuleManager();

	// zlib module
	i->NewGlobalTable("zlib");
	ILuaObject* table = i->GetGlobal("zlib");
		table->SetMember("Read",       compbuffer_read);
		table->SetMember("Compress",   zlib_compress);
		table->SetMember("Decompress", zlib_decompress);
		table->SetMember("version",    "1.1");
	table->UnReference();
	
	// b64 module
	i->NewGlobalTable("b64");
	table = i->GetGlobal("b64");
		table->SetMember("Encode", b64_encode);
		table->SetMember("Decode", b64_decode);
	table->UnReference();
	
	// CompBuffer object
	table = i->GetMetaTable(META_COMPBUFFER, TYPE_COMPBUFFER);
		table->SetMember("__gc",       compbuffer_gc);
		table->SetMember("__len",      compbuffer_len);		// compressed size
		table->SetMember("__tostring", b64_encode);			// base64 encode
		table->SetMember("__eq",       compbuffer_eq);		// equal to another CompBuffer? (checks sizes, checksums and contents)
		ILuaObject* __index = i->GetNewTable();
			__index->SetMember("Decompress",       zlib_decompress);
			__index->SetMember("Write",            compbuffer_write);		// write to file
			__index->SetMember("UncompressedSize", compbuffer_uncomp_size);
			__index->SetMember("Encode",           b64_encode);
			__index->SetMember("CompressedSize",   compbuffer_len);
			__index->SetMember("GetHeader",        compbuffer_getheader);
			__index->SetMember("SetHeader",        compbuffer_setheader);
		table->SetMember("__index", __index);
		__index->UnReference();
	table->UnReference();

	// Constants
	i->SetGlobal("Z_DEFAULT_STRATEGY", (float)0);
	i->SetGlobal("Z_FILTERED",         (float)1);
	i->SetGlobal("Z_HUFFMAN_ONLY",     (float)2);
	i->SetGlobal("Z_RLE",              (float)3);
	i->SetGlobal("Z_FIXED",            (float)4);

	// Create a new growable heap (starting at 4MB) for CompBuffers, temporary decompression space, etc.
	hHeap = HeapCreate(0, 4194304, 0);

	return 0;
}
extern "C" int __declspec(dllexport) gmod_close(lua_State*)
{
	HeapDestroy(hHeap);		// Free all used memory, all at once.
	hHeap = NULL;		// set to NULL because object __gc functions are called AFTER gmod_close.
	return 0;
}