/* zutil.c -- target dependent utility functions for the compression library
 * Copyright (C) 1995-2005 Jean-loup Gailly.
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

/* @(#) $Id$ */

// @!::
#include <windows.h>
#include "zutil.h"

#ifndef NO_DUMMY_DECL
struct internal_state	  {int dummy;}; /* for buggy compilers */
#endif

const char * const z_errmsg[10] = {
"need dictionary",	 /* Z_NEED_DICT	   2  */
"stream end",		  /* Z_STREAM_END	  1  */
"",					/* Z_OK			  0  */
"file error",		  /* Z_ERRNO		 (-1) */
"stream error",		/* Z_STREAM_ERROR  (-2) */
"data error",		  /* Z_DATA_ERROR	(-3) */
"insufficient memory", /* Z_MEM_ERROR	 (-4) */
"buffer error",		/* Z_BUF_ERROR	 (-5) */
"incompatible version",/* Z_VERSION_ERROR (-6) */
""};


const char * ZEXPORT zlibVersion()
{
	return ZLIB_VERSION;
}

// @!:: removed zlibCompileFlags

// @!:: used instead of memcpy and memset from CRT
void zmemcpy(dest, source, len)
    Bytef* dest;
    const Bytef* source;
    uInt  len;
{
    if (len == 0) return;
    do {
        *dest++ = *source++; /* ??? to be unrolled */
    } while (--len != 0);
}

int zmemcmp(s1, s2, len)
    const Bytef* s1;
    const Bytef* s2;
    uInt  len;
{
    uInt j;

    for (j = 0; j < len; j++) {
        if (s1[j] != s2[j]) return 2*(s1[j] > s2[j])-1;
    }
    return 0;
}

void zmemzero(dest, len)
    Bytef* dest;
    uInt  len;
{
    if (len == 0) return;
    do {
        *dest++ = 0;  /* ??? to be unrolled */
    } while (--len != 0);
}

/* exported to allow conversion of error code to string for compress() and
 * uncompress()
 */
const char* ZEXPORT zError(int err)
{
	return ERR_MSG(err);
}


// @!::
voidpf zcalloc(voidpf opaque, unsigned items, unsigned size)
{
	return HeapAlloc(opaque, 0, items * size);
}
void zcfree(voidpf opaque, voidpf ptr)
{
	HeapFree(opaque, 0, ptr);
}
